from enum import StrEnum
from typing import Annotated, Any

from pydantic import AfterValidator, BaseModel, Field
from pydantic_ai import ModelRetry


class RecommendedTool(BaseModel):
    recommended_tool_name: str
    recommended_tool_parameters: dict[str, Any]


class Step(BaseModel):
    step_number: Annotated[int, AfterValidator(lambda x: max(x, 1))] = Field(
        description="The sequential number of the step in the plan. Starts at 1."
    )
    instructions: str = Field(
        description=(
            "Clear, specific instructions for this step. Write these for a smaller AI model that "
            "needs explicit guidance and cannot infer context or handle ambiguity well. "
            "Be concrete, avoid complex reasoning, and include all necessary details."
        )
    )
    toolset_name: str = Field(
        default="none",
        description="The name of the toolset to fetch for the step. Set to 'none' if no toolset is needed.",
    )
    # recommended_tools: list[str] = Field(default_factory=list, description="The tool(s) to use for the step.")


class PlanStepStatus(StrEnum):
    PENDING = "pending"
    COMPLETED = "completed"
    FAILED = "failed"


class PlanStep(Step):
    status: PlanStepStatus = Field(default=PlanStepStatus.PENDING)
    # result: Any | None = None

    def __str__(self) -> str:
        res = f"<plan_step>\n<step_number>{self.step_number}</step_number>\n<instructions>{self.instructions}</instructions>\n<toolset_name>{self.toolset_name}</toolset_name>\n</plan_step>\n"
        # TODO: ADD RECOMMENDED TOOLS BACK LATER
        # TODO: ADD STATUS BACK LATER
        # TODO: ADD RESULT BACK LATER
        return res.strip()


class Plan(BaseModel):
    task: str = Field(description="The user's task to be executed.")
    steps: dict[int, PlanStep] = Field(default_factory=dict)

    def add_steps(self, steps: list[Step | PlanStep] | Step | PlanStep):
        for step in sorted(steps if isinstance(steps, list) else [steps], key=lambda x: x.step_number):
            self.steps[step.step_number] = PlanStep(**step.model_dump())

    # def mark_step_completed(self, step_number: int, result: Any | None = None):
    #     self.steps[step_number].status = PlanStepStatus.COMPLETED
    #     self.steps[step_number].result = result

    def get_step(self, step_number: int) -> PlanStep | None:
        return self.steps.get(step_number)

    def get_completed_steps(self) -> list[PlanStep]:
        return [step for step in self.steps.values() if step.status == PlanStepStatus.COMPLETED]

    def completed_steps_str(self) -> str:
        return (
            f"<completed_steps>\n"
            f"{'\n'.join(str(step) for step in self.get_completed_steps())}\n"
            f"</completed_steps>\n"
        )

    def get_next_pending_step(self) -> PlanStep | None:
        for step in self.steps.values():
            if step.status == PlanStepStatus.PENDING:
                return step
        return None

    def next_pending_step_str(self) -> str:
        return f"<next_pending_step>\n{str(self.get_next_pending_step())}\n</next_pending_step>\n"

    def __str__(self) -> str:
        return (
            f"<task>\n{self.task}\n</task>\n"
            f"<plan>\n"
            f"{'\n'.join(str(step) for step in self.steps.values())}\n"
            f"</plan>\n"
            "You may complete multiple steps in one go if you think it's appropriate."
        )


def present_plan(plan: str) -> str:
    """
    Present the complete plan to the user.
    This will not mention the tools/toolsets because the user is not technical.
    """
    return plan


def create_plan(task: str, steps: list[Step | PlanStep]) -> Plan:
    """Create the actual plan."""
    if not steps:
        raise ModelRetry("No steps provided. Please provide a list of steps.")
    plan = Plan(task=task)
    plan.add_steps(steps)
    return plan



from collections import defaultdict
from collections.abc import Callable
from dataclasses import replace
from typing import Concatenate, ParamSpec

from loguru import logger
from pydantic import BaseModel
from pydantic_ai.messages import (
    ModelMessage,
    ModelRequest,
    ModelRequestPart,
    ModelResponsePart,
    RetryPromptPart,
    ToolCallPart,
    ToolReturnPart,
)

EditFuncParams = ParamSpec("EditFuncParams")
EditFunc = Callable[
    Concatenate[ToolCallPart | ToolReturnPart, EditFuncParams], ToolCallPart | ToolReturnPart | None
]

def remove_retries(message_history: list[ModelMessage], keep_last_n: int = 1) -> list[ModelMessage]:
    if not isinstance(message_history[-1], ModelRequest):
        return message_history
    keep_last_n = min(len(message_history) - 1, max(0, keep_last_n))
    num_kept_retries: int = 0
    retry_tool_call_ids: list[str] = []
    successful_tools: set[str] = set()
    filtered_message_history: list[ModelMessage] = []
    for message in message_history[::-1]:
        filtered_parts: list[ModelRequestPart | ModelResponsePart] = []
        for part in message.parts:
            if isinstance(part, ToolReturnPart):
                successful_tools.add(part.tool_name)
            if isinstance(part, RetryPromptPart) and part.tool_name in successful_tools:
                if num_kept_retries < keep_last_n:
                    num_kept_retries += 1
                else:
                    retry_tool_call_ids.append(part.tool_call_id)
                    continue
            if isinstance(part, ToolCallPart) and part.tool_call_id in retry_tool_call_ids:
                continue
            filtered_parts.append(part)
        if filtered_parts:
            filtered_message_history.append(replace(message, parts=filtered_parts))
    logger.info(f"Skipped {len(retry_tool_call_ids)} retries")
    return filtered_message_history[::-1]


def remove_used_tools(
    message_history: list[ModelMessage], tool_names: list[str], lifespan: int | float = 3
) -> list[ModelMessage]:
    tools_to_remove: set[str] = set()
    filtered_message_history: list[ModelMessage] = []
    if isinstance(lifespan, float):
        lifespan = len(message_history) * lifespan
    num_skipped_parts: defaultdict[str, int] = defaultdict(int)
    for i, message in enumerate(message_history[::-1]):
        filtered_parts: list[ModelRequestPart | ModelResponsePart] = []
        for part in message.parts:
            if isinstance(part, ToolReturnPart) and part.tool_name in tool_names and i >= lifespan:
                tools_to_remove.add(part.tool_name)
            if (
                isinstance(part, (ToolCallPart, ToolReturnPart, RetryPromptPart))
                and part.tool_name in tools_to_remove
            ):
                num_skipped_parts[part.tool_name] += 1
                continue
            filtered_parts.append(part)
        if filtered_parts:
            filtered_message_history.append(replace(message, parts=filtered_parts))
    logger.info(f"Skipped parts: {dict(num_skipped_parts)}")
    return filtered_message_history[::-1]


class ToolEdit(BaseModel):
    tool_name: str
    edit_func: EditFunc[...]
    lifespan: int | float = 3


def edit_used_tools(
    message_history: list[ModelMessage], tools_to_edit_funcs: dict[str, ToolEdit]
) -> list[ModelMessage]:
    tools_to_edit: dict[str, EditFunc[...]] = {}
    filtered_message_history: list[ModelMessage] = []
    num_edited_parts: defaultdict[str, int] = defaultdict(int)
    for i, message in enumerate(message_history[::-1]):
        filtered_parts: list[ModelRequestPart | ModelResponsePart] = []
        for part in message.parts:
            if (
                isinstance(part, ToolReturnPart)
                and (tool_edit := tools_to_edit_funcs.get(part.tool_name)) is not None
            ):
                if isinstance(tool_edit.lifespan, float):
                    tool_edit.lifespan = len(message_history) * tool_edit.lifespan
                if i >= tool_edit.lifespan:
                    tools_to_edit[part.tool_name] = tool_edit.edit_func
            if isinstance(part, (ToolCallPart, ToolReturnPart)) and (
                (edit_func := tools_to_edit.get(part.tool_name)) is not None
            ):
                num_edited_parts[part.tool_name] += 1
                if (part := edit_func(part)) is None:
                    continue
            filtered_parts.append(part)
        if filtered_parts:
            filtered_message_history.append(replace(message, parts=filtered_parts))
    logger.info(f"Edited parts: {dict(num_edited_parts)}")
    return filtered_message_history[::-1]


def edit_tool_call_part(
    part: ToolCallPart | ToolReturnPart, content: str | None, thresh: int = 200
) -> ToolCallPart | ToolReturnPart | None:
    if isinstance(part, ToolReturnPart):
        return part
    if content is None:
        return None
    tool_args = part.args_as_json_str()
    if not tool_args or len(tool_args) < thresh:
        return part
    return ToolCallPart(
        tool_name=part.tool_name, args=content, tool_call_id=part.tool_call_id, part_kind=part.part_kind
    )


def edit_tool_result_part(
    part: ToolCallPart | ToolReturnPart, content: str | None, thresh: int = 200
) -> ToolCallPart | ToolReturnPart | None:
    if isinstance(part, ToolCallPart):
        return part
    if content is None:
        return None
    result_content = part.model_response_str()
    if len(result_content) < thresh:
        return part
    return ToolReturnPart(
        tool_name=part.tool_name,
        content=content,
        tool_call_id=part.tool_call_id,
        metadata=part.metadata,
        timestamp=part.timestamp,
        part_kind=part.part_kind,
    )



import inspect
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import Any

from dotenv import load_dotenv
from loguru import logger
from pydantic_ai import ModelRetry, RunContext
from pydantic_ai.tools import ToolDefinition

load_dotenv()


class Toolset:
    def __init__(
        self,
        name: str,
        tools: list[Callable[..., Any]],
        instructions: str | Path | Sequence[str | Path] | None = None,
        depends_on: list[str] | None = None,
    ):
        self.name = name
        if isinstance(instructions, (str, Path)):
            instructions = [instructions]
        self.instructions = (
            [
                instruction if not isinstance(instruction, Path) else instruction.read_text()
                for instruction in instructions
            ]
            if instructions
            else None
        )

        self.tools = tools
        self.depends_on = depends_on or []

    def describe(self, with_instructions: bool = False, with_tools: bool = True, full_doc: bool = True) -> str:
        res = f"<{self.name}>\n"
        if self.depends_on:
            res += f"Can't be fetched until the following toolsets are fetched: {self.depends_on}\n"
        if with_instructions and self.instructions:
            res += "<instructions>\n"
            res += "\n\n".join(self.instructions)
            res += "</instructions>\n"
        if with_tools and self.tools:
            res += "<tools>\n"
            for tool in self.tools:
                if full_doc:
                    doc = tool.__doc__ or "No description"
                else:
                    doc = (tool.__doc__ or "No description").split("\n\n")[0]  # Just summary before Args
                sig = inspect.signature(tool)
                res += f"{tool.__name__}{sig}: {doc.strip()}\n"
            res += "</tools>\n"
        res += f"</{self.name}>"
        return res


class AgentDeps:
    def __init__(self, toolset_registry: dict[str, Toolset] | None = None):
        self.toolset_registry: dict[str, Toolset] = toolset_registry or {}
        self.fetched_toolsets: set[str] = set()

    def add_toolsets(self, toolsets: list[Toolset] | Toolset):
        for toolset in toolsets if isinstance(toolsets, list) else [toolsets]:
            self.toolset_registry[toolset.name] = toolset


async def prepare_toolsets(ctx: RunContext[AgentDeps], tool_defs: list[ToolDefinition]) -> list[ToolDefinition]:
    """Prepare the toolsets: automatically skip tools from unfetched toolsets"""
    tools_to_skip: set[str] = set()
    for toolset_name, toolset in ctx.deps.toolset_registry.items():
        if toolset_name not in ctx.deps.fetched_toolsets or (
            toolset.depends_on and not all(dep in ctx.deps.fetched_toolsets for dep in toolset.depends_on)
        ):
            tools_to_skip.update(tool.__name__ for tool in toolset.tools)
    return [tool_def for tool_def in tool_defs if tool_def.name not in tools_to_skip]


async def list_available_toolsets(ctx: RunContext[AgentDeps]) -> str:
    """List available toolsets"""
    toolset_str = "<available_toolsets_for_fetching>\n"
    available_str = ""
    for toolset_name, toolset in ctx.deps.toolset_registry.items():
        if toolset_name not in ctx.deps.fetched_toolsets:
            available_str += f"{toolset.describe(with_instructions=False, full_doc=False)}\n"
    toolset_str += available_str
    toolset_str += "</available_toolsets_for_fetching>"
    return toolset_str


async def list_fetched_toolsets(ctx: RunContext[AgentDeps]) -> str:
    """List fetched toolsets"""
    toolset_str = "<fetched_toolsets>\n"
    fetched_str = ""
    for toolset_name in ctx.deps.fetched_toolsets:
        if toolset := ctx.deps.toolset_registry.get(toolset_name, None):
            fetched_str += f"{toolset.describe(with_instructions=True)}\n"
    toolset_str += fetched_str
    toolset_str += "</fetched_toolsets>"
    return toolset_str


async def fetch_toolset(ctx: RunContext[AgentDeps], toolset_name: str) -> str:
    """Fetch a toolset"""
    logger.info(f"Fetching `{toolset_name}` toolset from {list(ctx.deps.toolset_registry.keys())}")
    if ctx.deps.toolset_registry.get(toolset_name, None) is None:
        raise ModelRetry(f"The `{toolset_name}` toolset is not available.\n{await list_available_toolsets(ctx)}")
    if toolset_name in ctx.deps.fetched_toolsets:
        return f"You can already use tools from the `{toolset_name}` toolset."
    toolset = ctx.deps.toolset_registry[toolset_name]
    if toolset.depends_on and not all(dep in ctx.deps.fetched_toolsets for dep in toolset.depends_on):
        raise ModelRetry(
            f"The `{toolset_name}` toolset depends on the following toolsets: {toolset.depends_on}.\n"
            f"You need to fetch them first."
        )
    ctx.deps.fetched_toolsets.add(toolset_name)
    return f"You can now use tools from the `{toolset_name}` toolset."


async def _drop_toolset(ctx: RunContext[AgentDeps], toolset_name: str):
    """Drop a toolset"""
    logger.info(f"Dropping `{toolset_name}` toolset from {ctx.deps.fetched_toolsets}")
    if ctx.deps.toolset_registry.get(toolset_name, None) is None:
        raise ModelRetry(f"The `{toolset_name}` toolset is not available.\n{await list_available_toolsets(ctx)}")
    if toolset_name not in ctx.deps.fetched_toolsets:
        raise ModelRetry(f"You can't drop a toolset that you haven't fetched.\n{await list_fetched_toolsets(ctx)}")
    ctx.deps.fetched_toolsets.discard(toolset_name)


async def drop_toolsets(ctx: RunContext[AgentDeps], toolset_names: list[str]) -> str:
    """Drop one or more toolsets"""
    for toolset_name in toolset_names:
        await _drop_toolset(ctx, toolset_name)
    return (
        f"You can no longer use tools from the following toolsets: {', '.join(toolset_names)}.\n"
        f"You can fetch them again later if you need them by using the `fetch_toolset` tool."
    )



from functools import partial
from importlib.resources import files
from pathlib import Path

import logfire
from dotenv import load_dotenv
from pydantic_ai import Agent, ModelRetry
from pydantic_ai.messages import ModelMessage, ModelMessagesTypeAdapter
from pydantic_ai.models.openai import OpenAIModel
from pydantic_ai.providers.openrouter import OpenRouterProvider

from dreamai.basic_toolset import add_current_time_instructions, task_result, user_interaction
from dreamai.history_processors import remove_retries, remove_used_tools
from dreamai.toolsets import (
    AgentDeps,
    Toolset,
    drop_toolsets,
    fetch_toolset,
    list_available_toolsets,
    list_fetched_toolsets,
    prepare_toolsets,
)

load_dotenv()

logfire.configure(scrubbing=False)
logfire.instrument_pydantic_ai()
logfire.instrument_httpx(capture_all=True)


def add_numbers(numbers: list[int | float]) -> int | float:
    """Add numbers"""
    return sum(numbers)


def calculate_average(numbers: list[int | float]) -> int | float:
    """Calculate the average of numbers"""
    return sum(numbers) / len(numbers)


def guess_number(number: int) -> str:
    """Guess a number"""
    if number < 8:
        raise ModelRetry("Number is too low")
    return "correct!"


numerical_ops_tools = [add_numbers, calculate_average, guess_number]


model = OpenAIModel("deepseek/deepseek-chat-v3-0324", provider=OpenRouterProvider())
agent = Agent(
    model=model,
    instructions=[
        (files("dreamai.prompts") / "toolsets.md").read_text(),
        add_current_time_instructions,
        list_available_toolsets,
        list_fetched_toolsets,
    ],
    deps_type=AgentDeps,
    tools=[fetch_toolset, drop_toolsets] + numerical_ops_tools,
    output_type=[user_interaction, task_result],
    prepare_tools=prepare_toolsets,
    history_processors=[
        remove_retries,
        partial(remove_used_tools, tool_names=["guess_number"], lifespan=2),
        partial(remove_used_tools, tool_names=["fetch_toolset", "drop_toolsets"], lifespan=1),
    ],
    retries=10,
)

agent_deps = AgentDeps()
agent_deps.add_toolsets(Toolset(name="numerical_ops", tools=numerical_ops_tools))

if __name__ == "__main__":
    message_history_path = Path("message_history.json")
    message_history: list[ModelMessage] = (
        ModelMessagesTypeAdapter.validate_json(message_history_path.read_bytes())
        if message_history_path.exists()
        else []
    )
    task = input("> ") or "guess a number between 1 and 10 and tell me the result once you're correct"
    while True:
        result = agent.run_sync(user_prompt=task, deps=agent_deps, message_history=message_history)
        message_history = result.all_messages()
        task = input(f"{result.output}\n> ")
        if task.lower() in ["q", "quit", "exit"]:
            break
    message_history_path.write_bytes(result.all_messages_json())



from datetime import datetime

from pydantic import BaseModel, Field


def add_current_time_instructions() -> str:
    return f"<current_time>{datetime.now().astimezone().strftime('%Y-%m-%d %H:%M:%S')}</current_time>\n"


def user_interaction(message: str) -> str:
    """
    Interacts with the user. Could be:
    - A question
    - A progress update
    - An assumption made that needs to be validated
    - A request for clarification
    - Anything else needed from the user to proceed

    Args:
        message: The message to display to the user.
    """
    return message


class TaskResult(BaseModel):
    """
    A task result.
    """

    message: str = Field(description="The final response to the user.")


def task_result(message: str) -> TaskResult:
    """Returns the final response to the user."""
    return TaskResult(message=message)


def scratch_pad(thoughts: str):
    """Put your internal thoughts here in between toolsets and tools"""
    pass
